c============================================================================
c     ZSHMVM is a double precision Hermitian sparse matrix-vector multiplier 
c     which performs the matrix-vector product:
c
c                           A.x = y
c
c     where A is a sparse Hermitian matrix and x is a dense vector
c     
c     ZSHMVM(N,A,ROW,COL,NZ,X,LX,ANS)
c
c     Arguments:
c         N(input) INTEGER
c             number of rows in A
c         A(input) COMPLEX*16
c             nonzero elements of upper triangle of A
c         ROW(input) INTEGER
c             row indices of nonzero elements of upper triangle of A
c         COL(input) INTEGER
c             column indices of nonzero elements of upper triangel of A
c         NZ(input) INTEGER
c             number of nonzero elements of upper triangle of A
c         X(input) COMPLEX*16
c             vector x
c         LX(input) INTEGER
c             length of vector x
c         ANS(output) COMPLEX*16
c             the solution to the matrix-vector product A.x
c============================================================================

      subroutine ZSHMVM(N,A,ROW,COL,NZ,X,LX,ANS)

* declare variables

      implicit none
      integer N,NZ,LX,ROW(NZ),COL(NZ),i,k
      double complex A(NZ),X(LX),ANS(N)

* initialize answer vector

      do i=1,N
         ANS(i)=0
      end do

* calculate matrix-vector product

      k=1
      do i=1,N
         do while(k.le.NZ.AND.ROW(k).eq.i)
            ANS(i)=ANS(i)+A(k)*X(COL(k))
            if(ROW(k).ne.COL(k))
     *           ANS(COL(k))=ANS(COL(k))+conjg(A(k))*X(ROW(k))
            k=k+1
         end do
      end do
      return
      end

c============================================================================
c     ZSMVM is a double precision complex sparse matrix-vector multiplier 
c     which performs the matrix-vector product:
c
c                           A.x = y
c
c     where A is a complex sparse matrix(or its transpose, or Hermitian
c     conjugate) and x is a complex dense vector
c     
c     ZSMVM(TRANS,N,A,ROW,COL,NZ,X,LX,ANS)
c
c     Arguments:
c         TRANS(input) CHARACTER
c             N => A.x
c             T => A**T.x
c             C => A**H.x
c         N(input) INTEGER
c             number of rows in transformed A
c         A(input) COMPLEX*16
c             nonzero elements of A
c         ROW(input) INTEGER
c             row indices of nonzero elements of A
c         COL(input) INTEGER
c             column indices of nonzero elements of A
c         NZ(input) INTEGER
c             number of nonzero elements of A
c         X(input) COMPLEX*16
c             vector x
c         LX(input) INTEGER
c             length of vector x and number of columns in transformed A
c         ANS(output) COMPLEX*16
c             the solution to the matrix-vector product (transformed A).x
c============================================================================

      subroutine ZSMVM(TRANS,N,A,ROW,COL,NZ,X,LX,ANS)

* declare variables

      implicit none
      character TRANS
      integer N,NZ,LX,ROW(NZ),COL(NZ),i
      double complex A(NZ),X(LX),ANS(N)

* initialize answer vector

      do i=1,N
         ANS(i)=0
      end do

* calculate matrix-vector product y=A.x if TRANS=N, or y=A**H.x if TRANS=C

      if(TRANS.eq.'N')then
         do i=1,NZ
            ANS(ROW(i))=ANS(ROW(i))+A(i)*X(COL(i))
         end do
      else if(TRANS.eq.'T')then
         do i=1,NZ
            ANS(COL(i))=ANS(COL(i))+A(i)*X(ROW(i))
         end do
      else if(TRANS.eq.'C')then
         do i=1,NZ
            ANS(COL(i))=ANS(COL(i))+conjg(A(i))*X(ROW(i))
         end do
      end if
      return
      end

c============================================================================
c     DSSMVM is a double precision sparse symmetric matrix-vector multiplier
c     which performs the matrix-vector product:
c
c                           A.x = y
c
c     where A is a sparse symmetric matrix and x is a dense vector
c     
c     DSSMVM(N,A,ROW,COL,NZ,X,LX,ANS)
c
c     Arguments:
c         N(input) INTEGER
c             number of rows in A
c         A(input) DOUBLE PRECISION REAL
c             nonzero elements of upper triangle of A
c         ROW(input) INTEGER
c             row indices of nonzero elements of upper triangle of A
c         COL(input) INTEGER
c             column indices of nonzero elements of upper triangle of A
c         NZ(input) INTEGER
c             number of nonzero elements of upper triangle of A
c         X(input) DOUBLE PRECISION REAL
c             vector x
c         LX(input) INTEGER
c             length of vector x
c         ANS(output) DOUBLE PRECISION REAL
c             the solution to the matrix-vector product A.x
c============================================================================

      subroutine DSSMVM(N,A,ROW,COL,NZ,X,LX,ANS)

* declare variables

      implicit none
      integer N,NZ,LX,ROW(NZ),COL(NZ),i,k
      double precision A(NZ),X(LX),ANS(N)

* initialize answer vector

      do i=1,N
         ANS(i)=0
      end do

* calculate matrix-vector product

      k=1
      do i=1,N
         do while(ROW(k).eq.i)
            ANS(i)=ANS(i)+A(k)*X(COL(k))
            if(ROW(k).ne.COL(k))
     *           ANS(COL(k))=ANS(COL(k))+A(k)*X(ROW(k))
            k=k+1
         end do
      end do
      return
      end

c============================================================================
c     DSMVM is a double precision sparse matrix-vector multiplier which 
c     performs the matrix-vector product:
c
c                           A.x = y
c
c     where A is a sparse matrix(or its transpose) and x is a dense vector
c     
c     DSMVM(TRANS,N,A,ROW,COL,NZ,X,LX,ANS)
c
c     Arguments:
c         TRANS(input) CHARACTER
c             N => A.x
c             T => A**T.x
c         N(input) INTEGER
c             number of rows in transformed A
c         A(input) DOUBLE PRECISION REAL
c             nonzero elements of A
c         ROW(input) INTEGER
c             row indices of nonzero elements of A
c         COL(input) INTEGER
c             column indices of nonzero elements of A
c         NZ(input) INTEGER
c             number of nonzero elements of A
c         X(input) DOUBLE PRECISION REAL
c             vector x
c         LX(input) INTEGER
c             length of vector x and number of columns in transformed A
c         ANS(output) DOUBLE PRECISION REAL
c             the solution to the matrix-vector product (tranformed A).x
c============================================================================

      subroutine DSMVM(TRANS,N,A,ROW,COL,NZ,X,LX,ANS)

* declare variables

      implicit none
      character TRANS
      integer N,NZ,LX,ROW(NZ),COL(NZ),i
      double precision A(NZ),X(LX),ANS(N)

* initialize answer vector

      do i=1,N
         ANS(i)=0
      end do

* calculate matrix-vector product y=A.x if TRANS=N

      if(TRANS.eq.'N')then
         do i=1,NZ
            ANS(ROW(i))=ANS(ROW(i))+A(i)*X(COL(i))
         end do
      else if(TRANS.eq.'T')then
         do i=1,NZ
            ANS(COL(i))=ANS(COL(i))+A(i)*X(ROW(i))
         end do
      else if(TRANS.eq.'C')then
         do i=1,NZ
            ANS(COL(i))=ANS(COL(i))+A(i)*X(ROW(i))
         end do
      end if
      return
      end
