#include "sparsesd.h"
#include <float.h>
#define ZEROTEST(A) ((A) < DBL_EPSILON) /*macro for testing if equal to zero*/
#define REAL_C 0  /* 1 if C is real, 0 if C is imaginary
		   The compiler flag allows user to take advantage of
		   the type of partC, thus using less memory and a little
		   less time */

/****************************************************************************

  speigprob solves the semi-definite, Hermitian, generalized 
  eigenvalue problem: 
  
                       M.x=lambda.E.x
  
  for a sparse Hermitian matrix M, and positive semidefinite Hermitian
  matrix E.  Calculates a number of extreme eigenvalues using lanczos
  method.
  NOTE: compile with lanso, long, blas, lapack libraries, as well as
        spooles library (however there are bugs in spooles which could
	potentially cause problems??).

  Arguments:
      elslist *els:   contains matrix M, and innerproduct matrix E
      integer maxprs: number of ritz values to be calculated
      integer lanmax: maximum number of lanczos steps to be used
      integer lohi:   flag determining which extreme eigenvalues to calculate
                      = 0:  lowest eigenvalues
		      = 1:  highest eigenvaules
      integer *j:      number of lanczos steps actually taken
      integer *neig:  number of eigenvalues actually calculated
      doublereal *ritz:  sorted extreme eigenvalues calculated
      doublereal *bnd:   hold error bounds
      integer job:    flag determining what to calculate
                      = 0:  compute eigenvalues only;
                      = 1:  compute eigenvalues and vectors w/o constraints
                      = 2:  compute eigenvalues and vectors with constraints
      doublecomplex *vec:  holds eigenvectors if calculated
      integer msglvl: message level for lanczos
      integer *ierr:  error flag
      integer *info:  error flag
      int *info2:     error flag

*****************************************************************************/
void speigprob(elslist *els, integer maxprs, integer lanmax, integer lohi,
	       integer *j, integer *neig, doublereal *ritz, doublereal *bnd,
	       integer job, doublecomplex *vec, integer msglvl, 
	       integer *ierr, integer *info, int *info2)
{
  /* declare variables */

  int row, i;

  /* count number of zeros in els->inner */

  num = 0;
  *info = 0;
  for (i = 0; i < els->n; i++) {
    if (els->inner[i] == 0)
      ++num;
    if (els->inner[i] < 0) {
      *info = i;
      return;
    }
  }

  /* count number of elements in three block matrices */

  na=0;
  nb=0;
  nc=0;
  row=0;
  for (i = 0; i < els->j[els->n]; i++) {
    while (els->j[row+1] <= i)
      ++row;
    if (els->inner[row] != 0) {
      if (els->inner[els->i[i]] != 0)
	++na;
      else
	++nb;
    }
    else {
      if (els->inner[els->i[i]] != 0)
	++nb;
      else
	++nc;
    }
  }

  /* calculate eigenvalues and eigenvectors */

  speigprobA(els, maxprs, lanmax, lohi, j,
	     neig, ritz, bnd, job, vec, msglvl, ierr);

  *info2 = rc;
  return;
}

/*****************************************************************************

  speigproba  is called by speigprob to solve the semi-definite, Hermitian,
  generalized eigenvalue problem for sparse matrices, but only for the
  case where E is diagonal. 
  
  Arguments:
      elslist *els:   contains matrix M, and innerproduct matrix E
      integer maxprs: number of ritz values to be calculated
      integer lanmax: maximum number of lanczos steps to be used
      integer lohi:   flag determining which extreme eigenvalues to calculate
                      = 0:  lowest eigenvalues
		      = 1:  highest eigenvaules
      integer j:      number of lanczos steps actually taken
      integer *neig:  number of eigenvalues actually calculated
      doublereal *ritz:  sorted extreme eigenvalues calculated
      doublereal *bnd:   hold error bounds
      integer job:    flag determining what to calculate
                      = 0:  compute eigenvalues only;
                      = 1:  compute eigenvalues and vectors w/o constraints
                      = 2:  compute eigenvalues and vectors with constraints
      doublecomplex *vec:  holds eigenvectors if calculated
      integer msglvl: message level for lanczos
      integer *ierr:  error flag

*****************************************************************************/
void speigprobA(elslist *els, integer maxprs, integer lanmax, integer lohi,
		integer *j, integer *neig, doublereal *ritz, doublereal *bnd,
		integer job, doublecomplex *vec, integer msglvl,
		integer *ierr)
{
  /* declare variables and initialize values */

  int zcount, nzcount, i, k, kount, error, row, ai, ci, ar, cr,
    aknt, bknt, cknt;
  integer order;
  int *zeros = malloc(num * sizeof(int));
  int *nonzeros = malloc((els->n - num) * sizeof(int));
  parta = malloc(na * sizeof(complex_element));
  partac = malloc(na * sizeof(int));
  partar = malloc(na * sizeof(int));
  partb = malloc(nb * sizeof(complex_element));
  partbc = malloc(nb * sizeof(int));
  partbr = malloc(nb * sizeof(int));
#if REAL_C
  partcRealOnly = malloc(nc * sizeof(element)); /*real only uses half
						  as much memory*/
#else
  partc = malloc(nc * sizeof(complex_element));
#endif
  partcc = malloc(nc * sizeof(int));
  partcr = malloc(nc * sizeof(int));
  doublecomplex *q = malloc((els->n - num) * lanmax
			      * sizeof(doublecomplex));
  integer *list = malloc(els->n * sizeof(integer));
  ai = ci = kount = aknt = bknt = cknt = 0;
  row = ar = cr = -1;
  integer n = els->n;
  integer nz = els->j[n];
  doublereal kappa = 0;
  permuteflag = 1;
  order = n - num;
  doublereal condm = 1;
  integer retvec = 0;
  doublereal *transfer;
  error = 0;

  /* find zeros and nonzeros of els->inner, and calculate 1/sqrt(els->inner) */

  zcount = 0;
  nzcount = 0;
  for (i = 0; i < n; i++) {
    if (els->inner[i] == 0) {
      zeros[zcount] = i;
      ++zcount;
    }
    else {
      nonzeros[nzcount] = i;
      els->inner[i] = 1 / sqrt(els->inner[i]);
      ++nzcount;
    }
  }
 
  /* make list of indices for placing values in respective matrices */

  for(i = 0; i < n; i++) {
    if(ZEROTEST(els->inner[i]))
      list[i] = ci++;
    else
      list[i] = ai++;
  }
  if(n != (ai + ci))
    return;

  /* make block matrices A, B, and C from large matrix M */

  for(i = 0; i < nz; i++) {
    while(els->j[row+1] <= i) {
      ++row;
      if(ZEROTEST(els->inner[row]))
	++cr;
      else
	++ar;
    }
    /* part C */
    if(ZEROTEST(els->inner[row]) && ZEROTEST(els->inner[els->i[i]])) {
#if REAL_C
      if(els->c[i].im != 0) {
	printf("ERROR:  C is IMAGINARY, NOT JUST REAL!!!!!");/*error check*/
	return;
      }
      partcRealOnly[cknt] = els->c[i].re;
#else
      partc[cknt].re = els->c[i].re;
      partc[cknt].im = els->c[i].im;      
#endif
      partcr[cknt] = cr;
      partcc[cknt] = list[els->i[i]];
      cknt++;
    }
    /* part B dagger */
    else if(ZEROTEST(els->inner[row]) && !ZEROTEST(els->inner[els->i[i]])) {
      partb[bknt].re = els->c[i].re * els->inner[els->i[i]];
      partb[bknt].im = -els->c[i].im * els->inner[els->i[i]];
      partbr[bknt] = list[els->i[i]];
      partbc[bknt] = cr;
      bknt++;
    }
    /* part B */
    else if(!ZEROTEST(els->inner[row]) && ZEROTEST(els->inner[els->i[i]])) {
      partb[bknt].re = els->c[i].re * els->inner[row];
      partb[bknt].im = els->c[i].im * els->inner[row];
      partbr[bknt] = ar;
      partbc[bknt] = list[els->i[i]];
      bknt++;
    }
    /* part A */
    else {
      parta[aknt].re = els->c[i].re * els->inner[row] * els->inner[els->i[i]];
      parta[aknt].im = els->c[i].im * els->inner[row] * els->inner[els->i[i]];
      partar[aknt] = ar;
      partac[aknt] = list[els->i[i]];
      aknt++;
    }
  }
  /* free extra memory */

  free(els->c);
  free(els->i);
  free(els->j);
  els->c = NULL;
  els->i = NULL;
  els->j = NULL;
  free(list);

  /* factorize block matrix C */
  /* create and fill input matrix */
  InpMtx *mtxC = InpMtx_new();
#if REAL_C
  InpMtx_init(mtxC, INPMTX_BY_ROWS, SPOOLES_REAL, 0, 0);
  for(i = 0; i < nc; i++) {
    InpMtx_inputRealEntry(mtxC, partcr[i], partcc[i],
			     partcRealOnly[i]);
  }
#else
  InpMtx_init(mtxC, INPMTX_BY_ROWS, SPOOLES_COMPLEX, 0, 0);
  for(i = 0; i < nc; i++) {
    InpMtx_inputComplexEntry(mtxC, partcr[i], partcc[i],
			     partc[i].re, partc[i].im);
  }
#endif
  InpMtx_changeStorageMode(mtxC, INPMTX_BY_VECTORS);
  /* free extra memory */
#if REAL_C
  free(partcRealOnly);
#else
  free(partc);
#endif
  /* create bridge and set parameters */
  bridge = Bridge_new();
#if REAL_C
  Bridge_setMatrixParams(bridge, num, SPOOLES_REAL, SPOOLES_SYMMETRIC);
#else
  Bridge_setMatrixParams(bridge, num, SPOOLES_COMPLEX, SPOOLES_HERMITIAN);
#endif
  /* make sure pivoting is turned on */
  Bridge_setFactorParams(bridge, FRONTMTX_SPARSE_FRONTS, SPOOLES_PIVOTING,
 			 1000, .0001, NULL);
  /* setup bridge using my modified version of Bridge_setup */
  rc = Bridge_mySetup(bridge, mtxC);
  /* use bridge to factor */
  rc = Bridge_factor(bridge, mtxC, permuteflag, &error);

  /* call Lanczos to find eigenvalues and vectors */

  for(i = 0; i < order * lanmax; i++) {
    q[i].re = 0;
    q[i].im = 0;
  }
  /* note:  NEED TO CHANGE TO FORTRAN INDICES IN partar, partac,
            partbr, partbc FOR USE IN ZOP IN ZLANDR3!!!!*/
  for(i = 0; i < nb; i++) {
    partbr[i]++;
    partbc[i]++;
  }
  for(i = 0; i < na; i++) {
    partar[i]++;
    partac[i]++;
  }
  /* set return vector flag based on job # */
  if(job > 0)
    retvec = 1;
  ZLANDR3(&order, &lanmax, &maxprs, &lohi, &condm, &kappa, j, neig, ritz,
 	  bnd, &retvec, vec, q, ierr, &msglvl);

  /* calculate correct eigenvectors for job # */

  if(job == 2) {
    /* allocate memory for temporary workspace vectors */
    doublecomplex *xtmp = malloc(order * sizeof(doublecomplex));
    doublecomplex *ytmp = malloc(num * sizeof(doublecomplex));
    for(i = maxprs - 1; i > -1; i--) {
      /* initialize tmp vectors to zero */
      for(k = 0; k < order; k++) {
	xtmp[k].re = 0;
	xtmp[k].im = 0;
      }
      for(k = 0; k < num; k++) {
	ytmp[k].re = 0;
	ytmp[k].im = 0;
      }
      /* store X in bottom of xtmp */
      for(k = 0; k < order; k++) {
	xtmp[k].re = vec[(order * i) + k].re;
	xtmp[k].im = vec[(order * i) + k].im;
	/* resize eigenvectors X */
	vec[(order * i) + k].re = 
	  vec[(order * i) + k].re /  sqrt(els->inner[nonzeros[k]]);
 	vec[(order * i) + k].im = 
	  vec[(order * i) + k].im /  sqrt(els->inner[nonzeros[k]]);
      }
      /*===calculate eigenvectors Y in ytmp====*/
      char trans = 'C';
      ZSMVM(&trans, &num, partb, partbr, partbc, &nb, vec + (order * i),
	    &order, ytmp, 1);
      /* create and fill right hand side dense matrix */
      DenseMtx *mtxY = DenseMtx_new();
#if REAL_C
      /* Set up vector as two right hand sides, both real.  The first
         contains the real part and the second the imaginary. */
      DenseMtx_init(mtxY, SPOOLES_REAL, 0, 0, num, 2, 1, num);
      for(k = 0; k < num; k++) {
	DenseMtx_setRealEntry(mtxY, k, 0, ytmp[k].re);
	DenseMtx_setRealEntry(mtxY, k, 1, ytmp[k].re);
      }
#else
      DenseMtx_init(mtxY, SPOOLES_COMPLEX, 0, 0, num, 1, 1, num);
      for(k = 0; k < num; k++)
	DenseMtx_setComplexEntry(mtxY, k, 0, ytmp[k].re, ytmp[k].im);
#endif
      /* create and zero solution dense matrix */
      DenseMtx *mtxX = DenseMtx_new();
#if REAL_C
      DenseMtx_init(mtxX, SPOOLES_REAL, 0, 0, num, 2, 1, num);
#else
      DenseMtx_init(mtxX, SPOOLES_COMPLEX, 0, 0, num, 1, 1, num);
#endif
      DenseMtx_zero(mtxX);
      /* solve for solution using bridge */
      rc = Bridge_solve(bridge, permuteflag, mtxX, mtxY);
      /* store solution in ytmp */
      transfer = DenseMtx_entries(mtxX);
#if REAL_C
      /* break answer into the two parts, real and imaginary */
      for(k = 0; k < num; k++) {
	ytmp[i].re = transfer[i];
	ytmp[i].im = transfer[num + i];
      }
#else
      for(k = 0; k < num; k++) {
	ytmp[k].re = transfer[2 * k];
	ytmp[k].im = transfer[(2 * k) + 1];
      }
#endif
      /* sort X and Y back into M in the correct n */
      for(k = 0; k < num; k++) {
	vec[n * i + zeros[k]].re = -ytmp[k].re;
	vec[n * i + zeros[k]].im = -ytmp[k].im;
      }
      for(k = 0; k < order; k++) {
	vec[n * i + nonzeros[k]].re = xtmp[k].re;
	vec[n * i + nonzeros[k]].im = xtmp[k].im;
      }
      DenseMtx_free(mtxY);
      DenseMtx_free(mtxX);
    }
    free(xtmp);
    free(ytmp);
  }

  /* free extra memory */

  InpMtx_free(mtxC);
  Bridge_free(bridge);
  free(parta);
  free(partb);
  free(partar);
  free(partac);
  free(partbr);
  free(partbc);
  free(partcr);
  free(partcc);
  free(q);
  free(zeros);
  free(nonzeros);

  return;
}

/***************************************************************************

   ZOP is the linear Hermitian operator supplied to the complex lanczos
   routine, zlandr3; it calculates P=(A-B.Cinverse.Bdagger).Q

   Arguments:
       integer *n:        length of P, Q, and R
       doublecomplex *P:  unused in this particular operator
       doublecomplex *Q:  lanczos vector being acted on by operator
       doublecomplex *R:  result of operator acting on lanczos vector

****************************************************************************/
extern fsub ZOP(integer *n, doublecomplex *P, doublecomplex *Q,
	 	doublecomplex *R)
{
  /* declare variables */

  int i;
  char trans = 'C';
  integer incr = 1;
  doublereal *transfer;
  doublecomplex konst = {-1, 0};
  doublecomplex *tmp = malloc((*n + num) * sizeof(doublecomplex));

  /* calculate tmp = Bdagger.Q */

  ZSMVM(&trans, &num, partb, partbr, partbc, &nb, Q, n, tmp, 1);
  
  /*=========SPOOLES STUFF to calculate tmp = Cinverse.tmp========*/
  /* create and fill right hand side dense matrix */
  DenseMtx *mtxY = DenseMtx_new();
#if REAL_C
  /* Set up vector as two right hand sides, both real.  The first
     contains the real part and the second the imaginary. */
  DenseMtx_init(mtxY, SPOOLES_REAL, 0, 0, num, 2, 1, num);
  for(i = 0; i < num; i++) {
    DenseMtx_setRealEntry(mtxY, i, 0, tmp[i].re);
    DenseMtx_setRealEntry(mtxY, i, 1, tmp[i].im);
  }
#else
  DenseMtx_init(mtxY, SPOOLES_COMPLEX, 0, 0, num, 1, 1, num);
  for(i = 0; i < num; i++)
    DenseMtx_setComplexEntry(mtxY, i, 0, tmp[i].re, tmp[i].im);
#endif
  /* create and zero solution dense matrix */
  DenseMtx *mtxX = DenseMtx_new();
#if REAL_C
  DenseMtx_init(mtxX, SPOOLES_REAL, 0, 0, num, 2, 1, num);
#else
  DenseMtx_init(mtxX, SPOOLES_COMPLEX, 0, 0, num, 1, 1, num);
#endif
  DenseMtx_zero(mtxX);
  /* solve for solution using bridge */
  rc = Bridge_solve(bridge, permuteflag, mtxX, mtxY);
  /* store solution in tmp */
  transfer = DenseMtx_entries(mtxX);
#if REAL_C
  /* break answer into the two parts, real and imaginary */
  for(i = 0; i < num; i++) {
    tmp[i].re = transfer[i];
    tmp[i].im = transfer[num + i];
  }
#else
  for(i = 0; i < num; i++) {
    tmp[i].re = transfer[2 * i];
    tmp[i].im = transfer[(2 * i) + 1];
  }
#endif

  /* calculate R = B.tmp */

  trans = 'N';
  ZSMVM(&trans, n, partb, partbr, partbc, &nb, tmp, &num, R, 1);

  /* calculate tmp = A.Q */

  ZSHMVM(n, parta, partar, partac, &na, Q, n, tmp);

  /* calculate R = tmp - R */

  ZAXPY(n, &konst, tmp, &incr, R, &incr); 


  /* note: zaxpy is R = R - tmp, so must multiply by -1 */
  for(i = 0; i < *n; i++) {
    R[i].re = -R[i].re;
    R[i].im = -R[i].im;
  }

  /* free extra memory */

  DenseMtx_free(mtxY);
  DenseMtx_free(mtxX);
  free(tmp);
  return SUBRETURN;
}

/*************************************************************************

   ZOPM calculates the inner product in the lanczos routine, zlandr3, for
   double precision complex vectors; copies input vector to output vector

   Arguments:
       integer *n:        length of x and y
       doublecomplex *x:  input vector
       doublecomplex *y:  output vector

**************************************************************************/
extern fsub ZOPM(integer *n, doublecomplex *x, doublecomplex *y)
{
  memcpy(y, x, (*n) * sizeof(doublecomplex));
  return SUBRETURN;
}

/**************************************************************************

   junk function so code will compile

**************************************************************************/
extern fsub OP(integer *n, doublereal *P, doublereal *Q,
		doublereal *R)
{
  return SUBRETURN;
}

/**************************************************************************

   junk fuction so code will compile 

**************************************************************************/
extern fsub OPM(integer *n, doublereal *x, doublereal *y)
{
  memcpy(y, x, (*n) * sizeof(doublereal));
  return SUBRETURN;
}
